program EIKONAL


!==================================================================

!  If an input trace has a non-zero kill_val, then the eikonal solver
!  will not calculate a velocity for that triangle. Also, if the time-
!  delay value is negative, (they should all be positive) then the 
!  program will also kill that triangle. A good test is that the "number of
!  nonzero samples =" from:
!
!  Window3d < lb-stations-poly-nodes.H n1=1 f1=4 | Attr
!
!  is less than or equal to the same test on 'eikonal.H'

!==================================================================

use sep
use eikonal_mod
implicit none

integer					:: n1, n2, ntri, ntraces, nfreq, ff, nnodes, rec_l1, rec_p1, rec_l2, rec_p2
integer					:: nkeys, rec_point_key, rec_line_key, it, in, trace, found, n3out, killed1, killed2

real					:: freq, ofreq, dfreq, td(3)
real, dimension(:,:), allocatable	:: polys, output, all_timedelays, time_hdr, tmp, eik_out
real, dimension(:), allocatable		:: timedelay

 call sep_init()

 call from_history("n1",nfreq,1)
 call from_history("n2",ntraces,1)
 call from_history("o1",ofreq,1.0)
 call from_history("d1",dfreq,1.0)

 call from_aux("xylpk","n3",ntri,1)
 call from_aux("xylpk","n2",nnodes,3)
 call from_aux("hdr","n1",nkeys,16)

 call from_param("freq",freq,1.0)
 call from_param("rec_line_key",rec_line_key,3)
 call from_param("rec_point_key",rec_point_key,4)

 allocate( all_timedelays(nfreq,ntraces), timedelay(ntraces))
 allocate( polys(5,nnodes), output(4,nnodes), tmp(3,nnodes), eik_out(3,nnodes), time_hdr(nkeys,ntraces) )
 call sep_read(all_timedelays)
 call sep_read(time_hdr,'hdr')

! Grab a single frequency slice from the timedelays
ff = (freq - ofreq)/dfreq + 1
timedelay = all_timedelays(ff,:)

n3out=0
killed1=0
killed2=0


do it=1,ntri
	td = 0.0
        found=0

	! Read info for a single triangle
	call sep_read(polys,'xylpk')

	! Skip the triangles with killed nodes
	if (polys(5,1) > 0) then
		output = 1 ! (when ALL of output equals one, then the poly program will skip plotting it)
		call sep_write(output)
		killed1 = killed1 + 1
		cycle
	end if

	! If all nodes on a triangle are considered "live" then ...
	do in=1,nnodes

		! Get node rec_line and rec_point for node
		rec_l1 = polys(3,in)
		rec_p1 = polys(4,in)

		! Find trace that matches node (based on rec_line and rec_point)
		do trace=1, ntraces

			rec_l2 = time_hdr(rec_line_key,trace)
			rec_p2 = time_hdr(rec_point_key,trace)

			! But dont even consider trace if timedelay is < 0
                        if (timedelay(trace).le.0.) cycle

			! If a matching trace is found ...
			if ( (rec_l1==rec_l2) .and. (rec_p1==rec_p2) )then
				! then grab the time delay value
				td(in) = timedelay(trace)
				found = found + 1
				exit
			end if
		end do

		! Build (x, y, td) for this node
		tmp(1:2,in) = polys(1:2,in)	! (x,y) for a node
		tmp(3,in) = td(in)		! time delay (td) for a node
	end do

        if (found.ne.3) then
		output = 1 ! (when ALL of output equals one, then the poly program will skip plotting it)
		call sep_write(output)
		killed2 = killed2 + 1
		cycle		
	end if

	! Compute the eikonal plane solution
	call eikonal_solver(tmp,eik_out)
	output(4,:)=0
	output(1:3,:)=eik_out

	call sep_write(output)
        n3out=n3out+1
	
	if (exist_file('tmp')) then
		call sep_write(tmp, 'tmp')
	end if

end do


write(0,*) " NUMBER KILLED before:  ", killed1
write(0,*) " NUMBER KILLED (neg td):  ", killed2
write(0,*) " NUMBER FOUND:  ", n3out

 if (exist_file('tmp')) then
   call to_history("n1",4, 'tmp')
   call to_history("n2",3, 'tmp')
   call to_history("d1",1.0, 'tmp')
   call to_history("d2",1.0, 'tmp')
   call to_history("o1",0.0, 'tmp')
   call to_history("o2",0.0, 'tmp')
   call to_history("label1",'(x, y, velocity [km/s]), kill_val', 'tmp')
   call to_history("label2",'Node', 'tmp')
 end if

 call to_history("n1",4)
 call to_history("n2",nnodes)
 call to_history("n3",ntri)
 call to_history("o1",1)
 call to_history("d1",1)
 call to_history("label1",'(x, y, velocity [km/s]), kill_val)' )
 call to_history("label2",'Node' )
 call to_history("label3",'Triangle' )

 call sep_close()
deallocate(all_timedelays, timedelay, polys, output, time_hdr)
end program


